from __future__ import annotations

import asyncio
import typing
from collections import abc

import psycopg2
import pydantic
import sprockets_postgres

import imbi.opensearch.project
from imbi import common, errors
from imbi.endpoints import base


class Integration(pydantic.BaseModel):
    """Extracts selected columns from v1.integrations"""
    name: str


class NotificationRule(pydantic.BaseModel):
    """Extracts selected columns from v1.notification_rules"""
    fact_type_id: int
    pattern: common.JsonPointer


class NotificationFilter(pydantic.BaseModel):
    """Extracts selected columns from v1.notification_filters"""
    name: str
    pattern: common.JsonPointer
    operation: str
    value: str
    action: str


class Notification(pydantic.BaseModel):
    """Model of a v1.notifications row and related entities"""
    integration: Integration
    id_pattern: common.JsonPointer
    verification_token: typing.Union[str, None] = None
    documentation: typing.Union[str, None] = None
    default_action: str
    rules: typing.List[NotificationRule] = pydantic.Field(default_factory=list)
    filters: typing.List[NotificationFilter] = pydantic.Field(
        default_factory=list)


class ProjectInfo(pydantic.BaseModel):
    """Extracts selected columns from v1.projects"""
    id: int
    project_type_id: int


T: typing.TypeAlias = typing.TypeVar('T')
Pair: typing.TypeAlias = tuple[T, T]


class ProcessingHandler(base.RequestHandler):
    """Handle incoming notifications from integrated applications.

    This handler is a little strange in that it requires that the last
    portion of the path match the request method. This makes the
    configuration of a webhook clearer since you are explicitly stating
    what the expected HTTP method and information passing mechanism is.
    Most webhooks will use the `.../post` variant.  The other path
    elements identify the source integration and notification type.

    Request processing will only fail for operational or configuration
    reasons. If an incoming notification does not match an existing
    project or fails to generate an update list, then the handler will
    exit successfully. This is by design.

    If the request targets a non-existent integration or notification,
    then a 404 error will be returned. This is essentially an incorrect
    path.

    """
    async def prepare(self) -> None:
        # make sure that the appropriate handler is being invoked
        # based on the last portion of the path
        await super().prepare()
        method_from_path = self.path_kwargs['action'].upper()
        if self.request.method.upper() != method_from_path:
            raise errors.BadRequest(
                'request method %r does not match expected %r',
                self.request.method.upper(), method_from_path)

    async def _process_notification(self, integration_name, notification_name,
                                    body) -> None:
        """Process the incoming notification.

        The request processing methods (eg, `post`, `put`, etc.) are
        responsible for gathering the body of the notification from
        the request and calling this method to do the actual processing.

        """
        self.logger.name = '.'.join(
            [__package__, integration_name, notification_name])
        notification = await self._get_notification(integration_name,
                                                    notification_name)
        if not self._evaluate_filters(notification, body):
            return

        self.logger.info('processing notification %s/%s default %r',
                         integration_name, notification_name,
                         notification.default_action)
        search_index = imbi.opensearch.project.ProjectIndex(self.application)
        for project in await self._get_projects(notification, body):
            self.logger.debug('checking for updates on imbi project id=%s',
                              project.id)
            updates = await self._gather_updates(project, notification, body)
            if updates:
                self.logger.info('updating %s facts on imbi project id=%s',
                                 len(updates), project.id)
                await self._update_facts(project, updates)
                await search_index.index_document_by_id(project.id)

    async def get(self, *, integration_name: str, notification_name: str,
                  **_kwargs: str) -> None:
        body: dict[str, str | list[str] | None] = {}
        for arg, values in self.request.query_arguments.items():
            # NB ..../get?foo results in {'foo': [b'']} and NOT
            # {'foo': []} as one might otherwise expect
            if (len(values) == 0 or len(values) == 1
                    and values[0] == b''):  # pragma: no branch
                body[arg] = None
            elif len(values) == 1:
                body[arg] = values[0].decode('utf-8')
            else:
                raise errors.UnprocessableEntity(
                    'multiple query arguments with same name %r', arg)
        await self._process_notification(integration_name, notification_name,
                                         body)

    async def post(self, *, integration_name: str, notification_name: str,
                   **_kwargs: str) -> None:
        await self._process_notification(integration_name, notification_name,
                                         self.get_request_body())

    put = post

    async def _update_facts(self, project, updates) -> None:

        async with self.application.postgres_connector(
                on_duration=self.on_postgres_timing) as connector:
            for fact_type_id, value in updates.items():
                try:
                    await connector.execute(
                        'INSERT INTO v1.project_facts'
                        '            (project_id, fact_type_id, value,'
                        '             recorded_at, recorded_by)'
                        '     VALUES (%(project_id)s, %(fact_type_id)s,'
                        '             %(value)s, CURRENT_TIMESTAMP,'
                        '             %(username)s)'
                        '         ON CONFLICT (project_id, fact_type_id)'
                        '         DO UPDATE'
                        '        SET value = %(value)s,'
                        '            recorded_at = CURRENT_TIMESTAMP,'
                        '            recorded_by = %(username)s',
                        parameters={
                            'project_id': project.id,
                            'fact_type_id': fact_type_id,
                            'value': value,
                            'username': 'system'
                        },
                        metric_name='upsert-project-fact')
                except (asyncio.TimeoutError, psycopg2.Error) as error:
                    raise errors.InternalServerError(
                        'failed to update fact %s for project %s with value'
                        ' %r: %s', fact_type_id, project.id, value, error)

    async def _get_notification(self, integration_name: str,
                                notification_name: str) -> Notification:
        key = {
            'integration_name': integration_name,
            'notification_name': notification_name
        }
        queries = [
            self.postgres_execute(
                'SELECT name'
                '  FROM v1.integrations'
                ' WHERE name = %(integration_name)s',
                parameters=key,
                metric_name='fetch-integration'),
            self.postgres_execute(
                'SELECT id_pattern, verification_token, documentation,'
                '       default_action'
                '  FROM v1.integration_notifications'
                ' WHERE integration_name = %(integration_name)s'
                '   AND notification_name = %(notification_name)s',
                parameters=key,
                metric_name='fetch-notifications'),
            self.postgres_execute(
                'SELECT filter_name AS name, pattern, operation, value, action'
                '  FROM v1.notification_filters'
                ' WHERE integration_name = %(integration_name)s'
                '   AND notification_name = %(notification_name)s',
                parameters=key,
                metric_name='fetch-notification-filters'),
            self.postgres_execute(
                'SELECT fact_type_id, pattern'
                ' FROM v1.notification_rules'
                ' WHERE integration_name = %(integration_name)s'
                '   AND notification_name = %(notification_name)s',
                parameters=key,
                metric_name='fetch-notification-rules')
        ]
        results: tuple[sprockets_postgres.QueryResult, ...]
        results = await asyncio.gather(*queries)
        integration_data, notification_data, filters, rules = results

        if not integration_data.row:
            raise errors.ItemNotFound('Integration %r not found',
                                      integration_name)
        if not notification_data.row:
            raise errors.ItemNotFound('Notification %r not found',
                                      notification_name)

        integration = Integration.model_validate(integration_data.row)
        notification_data.row['integration'] = integration.model_dump()
        notification = Notification.model_validate(notification_data.row)
        notification.rules.extend(
            NotificationRule.model_validate(row) for row in rules.rows)
        notification.filters.extend(
            NotificationFilter.model_validate(row) for row in filters.rows)

        return notification

    def _coerce_value(
        self,
        constraint: str,
        notification_value: str | float | int | bool,
    ) -> Pair[int] | Pair[float] | Pair[bool] | Pair[str]:
        """Coerce the string constraint according to received value"""
        try:
            if isinstance(notification_value,
                          bool) and constraint.lower() in {'true', 'false'}:
                self.logger.debug('_coercing %r and %r to boolean', constraint,
                                  notification_value)
                return constraint.lower() == 'true', notification_value
            if isinstance(notification_value, float):
                self.logger.debug('_coercing %r and %r to float', constraint,
                                  notification_value)
                return float(constraint), notification_value
            if isinstance(notification_value, int) and constraint.isdigit():
                self.logger.debug('_coercing %r and %r to int', constraint,
                                  notification_value)
                return int(constraint, 10), notification_value
        except (TypeError, ValueError):
            self.logger.warning('failed to coerce %r to a %s', constraint,
                                notification_value.__class__.__name__)
        return constraint, notification_value

    def _evaluate_filters(self, notification: Notification, body) -> bool:
        """Evaluate the list of filters.

        If the default action is "process", then the notification
        should be processed unless there is a matching "ignore"
        filter.

        If the default action is "ignore", then the notification
        should be processed if there is at least one matching
        "process" filter and zero matching "ignore" filters.

        :returns: ``True`` if the notification should be processed
            and ``False`` otherwise

        """
        matches = []
        unspecified = object()
        for notification_filter in notification.filters:
            constraint = notification_filter.value
            value = notification_filter.pattern.resolve(body, unspecified)
            self.logger.debug('evaluating filter %s with %r',
                              notification_filter,
                              'unmatched' if value is unspecified else value)
            if value is not unspecified:
                constraint, value = self._coerce_value(constraint, value)
                if value == constraint:
                    if notification_filter.operation == '==':
                        self.logger.debug('matched %s -> %s',
                                          notification_filter.name,
                                          notification_filter.action)
                        matches.append((notification_filter, value))
                else:
                    if notification_filter.operation == '!=':
                        self.logger.debug('matched %s -> %s',
                                          notification_filter.name,
                                          notification_filter.action)
                        matches.append((notification_filter, value))

        should_process = None
        if matches:
            ignores = [(f, t) for f, t in matches if f.action == 'ignore']
            process = [(f, t) for f, t in matches if f.action == 'process']
            if ignores:
                should_process = False
                for f, v in ignores:
                    self.logger.info('ignoring %s found %r instead of %r',
                                     f.name, v, f.value)
            elif process:
                should_process = True
                for f, v in process:
                    self.logger.info('processing: %s matched %r', f.name, v)
        if should_process is None:
            self.logger.info('using default %s', notification.default_action)
            should_process = notification.default_action == 'process'
        return should_process

    async def _get_projects(self, notification: Notification,
                            body) -> abc.Iterable[ProjectInfo]:
        surrogate_project_id = notification.id_pattern.resolve(body, None)
        if surrogate_project_id is None:
            self.logger.warning('failed to find surrogate project id using %s',
                                notification.id_pattern)
            return []

        self.logger.debug('looking for project %s@%s', surrogate_project_id,
                          notification.integration.name)
        result = await self.postgres_execute(
            'SELECT p.id, p.project_type_id'
            '  FROM v1.project_identifiers AS i'
            '  JOIN v1.projects AS p ON (p.id = i.project_id)'
            ' WHERE i.external_id = %(external_id)s::TEXT'
            '   AND i.integration_name = %(integration_name)s',
            parameters={
                'external_id': surrogate_project_id,
                'integration_name': notification.integration.name,
            },
            metric_name='get-project-by-surrogate-id')
        if not result:
            self.logger.warning(
                'failed to find imbi project for %s@%s',
                surrogate_project_id,
                notification.integration.name,
            )
            return []
        return [ProjectInfo.model_validate(r) for r in result.rows]

    async def _gather_updates(self, project: ProjectInfo,
                              notification: Notification,
                              body) -> dict[int, str]:
        """Process the notification rules into a dict of updates

        :returns: a dictionary that maps from fact type ID to the new value

        """
        result = await self.postgres_execute(
            'SELECT id'
            '  FROM v1.project_fact_types'
            ' WHERE project_type_ids @> ARRAY[%s]', [project.project_type_id],
            metric_name='get-facts-types-for-notification')
        if not result:
            self.logger.debug('no project facts defined for project type %s',
                              project.project_type_id)
            return {}

        unspecified = object()
        updates = {}
        valid_fact_type_ids = {r['id'] for r in result}
        for rule in notification.rules:
            if rule.fact_type_id in valid_fact_type_ids:
                value = rule.pattern.resolve(body, unspecified)
                if value is not unspecified:
                    updates[rule.fact_type_id] = value
        return updates



from __future__ import annotations

from imbi.endpoints import base


class CollectionRequestHandler(base.CollectionRequestHandler):
    NAME = 'project-notes'

    ID_KEY = ['id', 'project_id']
    """Tell the framework what parts of the path to use as query parameters"""

    FIELDS = ['content', 'project_id']
    """Identifies properties in the POST request

    These can come from the request body or the path parameters and
    are the values available to `POST_SQL`.  Note that `username` is
    added automatically based on the current user.

    """

    ITEM_NAME = 'project-note'
    """Identifies the route name for a single item in the collection."""

    COLLECTION_SQL = ('SELECT id, content, project_id, created_by, updated_by'
                      '  FROM v1.project_notes'
                      ' WHERE project_id = %(project_id)s'
                      ' ORDER BY created_at DESC, id DESC')

    GET_SQL = ('SELECT id, content, created_by, project_id, updated_by'
               '  FROM v1.project_notes'
               ' WHERE id = %(id)s')

    POST_SQL = ('INSERT INTO v1.project_notes(project_id, content, created_by)'
                '     VALUES (%(project_id)s, %(content)s, %(username)s)'
                '  RETURNING *')


class RecordHandler(base.CRUDRequestHandler):
    NAME = 'project-note'
    ID_KEY = ['id', 'project_id']
    GET_SQL = ('SELECT id, content, created_by, project_id, updated_by'
               '  FROM v1.project_notes'
               ' WHERE id = %(id)s'
               '   AND project_id = %(project_id)s')
    PATCH_SQL = ('UPDATE v1.project_notes'
                 '   SET updated_by = %(username)s,'
                 '       updated_at = CURRENT_TIMESTAMP,'
                 '       content = %(content)s'
                 ' WHERE id = %(current_id)s'
                 '   AND project_id = %(current_project_id)s')
    DELETE_SQL = ('DELETE FROM v1.project_notes'
                  ' WHERE id = %(id)s'
                  '   AND project_id = %(project_id)s')



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ITEM_NAME = 'cookie-cutter'
    ID_KEY = 'name'
    FIELDS = ['name', 'description', 'type', 'project_type', 'url']
    TTL = 300

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT "name", created_at, created_by, last_modified_at,
               last_modified_by, description, "type", project_type_id, url
          FROM v1.cookie_cutters
         WHERE "name"=%(name)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'cookie-cutters'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT "name", "type", project_type_id, description, url
          FROM v1.cookie_cutters
         ORDER BY "name" ASC""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.cookie_cutters
                    ("name", created_by, project_type_id, "type",
                     description, url)
             VALUES (%(name)s, %(username)s, %(project_type_id)s, %(type)s,
                     %(description)s, %(url)s)
          RETURNING "name";""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'cookie-cutter'

    DELETE_SQL = 'DELETE FROM v1.cookie_cutters WHERE "name"=%(name)s'

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.cookie_cutters
           SET "name"=%(name)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               description=%(description)s,
               "type"=%(type)s,
               project_type_id=%(project_type_id)s,
               url=%(url)s
         WHERE "name"=%(current_name)s""")



"""
Interface for aggregating stats into Redis for single view of all metrics when
polling.

"""
import contextlib
import os
import socket
import time
import typing

import aioredis

DEFAULT_POOL_SIZE = 10


class Stats:
    """Class used for aggregating stats to be exposed to an external metric
    collector.

    """
    def __init__(self, client: aioredis.Redis):
        """Create a new instance of the Stats class"""
        self._client = client
        self._hostname = socket.gethostname()

    async def decr(self, tags: dict, value: int = 1) -> None:
        """Decrement a counter

        :param tags: The key tags to decrement
        :param value: The value to decrement by

        """
        await self._client.decrby(self._compose_key('c', tags), value)

    async def incr(self, tags: dict, value: int = 1) -> None:
        """Increment a counter

        :param tags: The key tags to increment
        :param value: The value to increment by

        """
        await self._client.incrby(self._compose_key('c', tags), value)

    async def add_duration(self, tags: dict, value: float) -> None:
        """Add a duration for the specified key

        :param tags: The key tags to record the timing for
        :param value: The value of the duration

        """
        await self._client.lpush(self._compose_key('d', tags), value)

    @contextlib.asynccontextmanager
    async def track_duration(self, tags: dict) -> typing.AsyncContextManager:
        """Context manager that sets a value with the duration of time that it
        takes to execute whatever it is wrapping.

        :param tags: The key tags to track the duration of

        """
        start_time = time.time()
        try:
            yield
        finally:
            await self.add_duration(tags, time.time() - start_time)

    async def counters(self,
                       all_hosts: bool = False,
                       flush: bool = False) -> dict[str, int]:
        """Return a dict of counters and their values

        :param all_hosts: Process counters for all hosts
        :param flush: Flush existing values

        """
        counters = {}
        keys = await self._client.keys(
            'c:*' if all_hosts else f'c:*hostname={self._hostname}*')
        if keys:
            values = await self._client.mget(*keys)
            for offset, key in enumerate(keys):
                counters[key[2:].decode('utf-8')] = int(values[offset])
            if flush:
                await self._client.delete(*keys)
        return counters

    async def durations(self,
                        all_hosts: bool = False,
                        flush: bool = False) \
            -> dict[str, list[float]]:
        """Return a dict of durations and their values

        :param all_hosts: Process durations for all hosts
        :param flush: Flush existing values

        """
        durations = {}
        keys = await self._client.keys(
            'd:*' if all_hosts else f'd:*hostname={self._hostname}*')
        for key in keys:
            durations[key[2:].decode('utf-8')] = sorted(
                float(v) for v in await self._client.lrange(key, 0, -1))
            if flush:
                await self._client.delete(key)
        return durations

    def _compose_key(self, prefix: str, tags: dict) -> str:
        """Create the deterministically sorted compound key that is used
        to ensure we're always incrementing or adding to the same set for
        the same set of values.

        :param prefix: The key prefix
        :param tags: The key tags that will be used to compose the key

        """
        parts = [f'hostname={self._hostname}']
        for key, value in tags.items():
            parts.append(f'{key}={value}')
        return f'{prefix}:' + ':'.join(value for value in sorted(parts))


async def create(redis_url):
    """Create a new instance of the Stats class

    :param str redis_url: The Redis URL to use for stats
    :rtype: Stats


    """
    redis_client = await aioredis.create_pool(
        redis_url,
        maxsize=int(os.environ.get('STATS_POOL_SIZE', DEFAULT_POOL_SIZE)))
    return Stats(redis_client)



from __future__ import annotations

import base64
import datetime
import decimal
import typing

import dateutil.parser
import jsonpointer
import pydantic
import yarl
from ietfparse import datastructures
from pydantic_core import core_schema


def coerce_project_fact_values(
    rows: list[dict[str, typing.Any]]
) -> list[dict[str, None | bool | decimal.Decimal | int | str]]:
    """Shared method to coerce the values to the correct data types in the
    output for project facts

    """
    for row in rows:
        row['value'] = coerce_project_fact(row['data_type'], row['value'])
        del row['data_type']

    return rows


def coerce_project_fact(
        data_type: str,
        value: typing.Any) -> None | bool | decimal.Decimal | int | str:
    """Coerce a single fact based on its data type."""
    if not value and value != 0:
        value = None
    elif data_type == 'boolean':
        value = 'false' if value is None else str(value).lower()
        if value in ('y', 'yes', 't', 'true', 'on', '1'):
            value = True
        elif value in ('n', 'no', 'f', 'false', 'off', '0'):
            value = False
        else:
            raise ValueError(f'{value!r} is not a valid Boolean')
    elif data_type == 'decimal':
        try:
            value = decimal.Decimal(value)
        except decimal.InvalidOperation as error:
            raise ValueError(f'{value!r} is not a valid decimal: {error}')
    elif data_type == 'integer':
        try:
            value = int(value)
        except TypeError as error:
            raise ValueError(f'{value!r} is not a valid integer: {error}')
    elif data_type == 'date':
        value = dateutil.parser.parse(str(value))
        value = value.date().isoformat()
    elif data_type == 'timestamp':
        value = dateutil.parser.parse(str(value))
        if value.tzinfo is None:
            value = value.replace(tzinfo=datetime.timezone.utc)
        value = value.isoformat()
    elif data_type == 'string':
        value = str(value)
    else:
        raise ValueError(f'{data_type!r} is not a known fact type')

    return value


def urlsafe_padded_b64decode(data: str) -> bytes:
    """Base64 decode 'data' after padding with '=' if needed"""
    bin_data = data.encode('ascii')
    data_len = len(bin_data)
    pad_len = data_len % 4
    if pad_len:
        bin_data = bin_data.ljust(data_len + pad_len, b'=')
    return base64.urlsafe_b64decode(bin_data)


def build_link_header(target: yarl.URL, rel: str, **parameters: str) -> str:
    all_params = [('rel', rel)]
    all_params.extend(parameters.items())
    return str(datastructures.LinkHeader(str(target), all_params))


def validated_jsonpointer(
        value: str | jsonpointer.JsonPointer) -> jsonpointer.JsonPointer:
    """Convert a string to a valid jsonpointer.JsonPointer

    The JsonPointer initializer does not fail when given an empty
    string. This function raises `TypeError` and `ValueError` when
    appropriate.

    """
    if isinstance(value, jsonpointer.JsonPointer):
        return value
    if not isinstance(value, str):
        raise TypeError('Value must be a string')
    try:
        pointer = jsonpointer.JsonPointer(value)
    except jsonpointer.JsonPointerException as error:
        raise ValueError(error.args[0]) from error
    else:
        if not pointer.parts:
            raise ValueError('JSON pointer must not be empty')
        return jsonpointer.JsonPointer(value)


def _jptr_core_schema(*_) -> core_schema.CoreSchema:
    """Small helper for use with pydantic.GetPydanticSchema"""
    validator = core_schema.no_info_plain_validator_function(
        validated_jsonpointer)
    return core_schema.json_or_python_schema(json_schema=validator,
                                             python_schema=validator)


def _jptr_json_schema(*_) -> pydantic.json_schema.JsonSchemaValue:
    """Small helper for use with pydantic.GetPydanticSchema"""
    return {'type': 'string', 'format': 'json-pointer'}


JsonPointer = typing.Annotated[
    jsonpointer.JsonPointer,
    pydantic.GetPydanticSchema(_jptr_core_schema, _jptr_json_schema)]
"""Pydantic capable version of a jsonpointer.JsonPointer"""



"""
Base Request Handlers

"""
from __future__ import annotations

import asyncio
import base64
import contextlib
import dataclasses
import datetime
import logging
import operator
import uuid
from email import utils

import jsonpatch
import jsonpointer
import openapi_core.casting.schemas.exceptions
import problemdetails
import pydantic
import sprockets.mixins.mediatype.content
import sprockets_postgres as postgres
import typing_extensions as typing
import umsgpack
import yarl
from ietfparse import datastructures
from openapi_core.deserializing.exceptions import DeserializeError
from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.paths.exceptions import \
    OperationNotFound, PathNotFound
from openapi_core.unmarshalling.schemas.exceptions import ValidateError
from openapi_core.validation.exceptions import InvalidSecurity
from sprockets.http import mixins
from tornado import httputil, web

from imbi import common, cors, errors, session, user, version

LOGGER = logging.getLogger(__name__)


def require_permission(permission):
    """Decorator function for requiring a permission string for an endpoint

    :param str permission: The permission string to require
    :raises: problemdetails.Problem

    """
    def _require_permission(f):
        def wrapped(self, *args, **kwargs):
            """Inner-wrapping of the decorator that performs the logic"""
            if not self._current_user or \
                    not self._current_user.has_permission(permission):
                if self._respond_with_html:
                    return self.render(
                        'index.html',
                        javascript_url=self.application.settings.get(
                            'javascript_url'))
                raise errors.Forbidden(
                    '%s(%s) does not have the "%s" permission',
                    self._current_user.username, self._current_user.user_type,
                    permission)
            return f(self, *args, **kwargs)

        return wrapped

    return _require_permission


class RequestHandler(cors.CORSMixin, postgres.RequestHandlerMixin,
                     mixins.ErrorLogger, problemdetails.ErrorWriter,
                     sprockets.mixins.mediatype.content.ContentMixin,
                     web.RequestHandler):
    """Base RequestHandler class used for recipients and subscribers."""

    APPLICATION_JSON = 'application/json'
    TEXT_HTML = 'text/html'
    NAME = 'Base'
    ITEM_NAME = ''

    def __init__(self, application, request: httputil.HTTPServerRequest,
                 **kwargs):
        super().__init__(application, request, **kwargs)
        self.logger = logging.getLogger(f'imbi.endpoints.{self.NAME}')
        self.session: typing.Optional[session.Session] = None
        self._current_user: typing.Optional[user.User] = None
        self._links = {}

    async def prepare(self) -> None:
        """Prepare the request handler for the request. If the application
        is not ready return a ``503`` error.

        Checks for a session cookie and if present, loads the session into
        the current user and authenticates it. If authentication fails,
        the current user and cookie is cleared.

        """
        if not self.application.ready_to_serve:
            return self.send_error(503, reason='Application not ready')
        self.session = session.Session(self)
        await self.session.initialize()
        self._current_user = await self.get_current_user()
        future = super().prepare()
        if asyncio.isfuture(future) or asyncio.iscoroutine(future):
            await future

    def on_finish(self) -> None:
        """Invoked after a request has completed"""
        super().on_finish()
        self.application.loop.add_callback(
            self.application.stats.incr, {
                'key': 'http_requests',
                'endpoint': self.NAME,
                'method': self.request.method,
                'status': self.get_status()
            })

        self.application.loop.add_callback(
            self.application.stats.add_duration, {
                'key': 'http_request_duration',
                'endpoint': self.NAME,
                'method': self.request.method,
                'status': self.get_status()
            }, self.request.request_time())

    def compute_etag(self) -> None:
        """Override Tornado's built-in ETag generation"""
        return None

    async def get_current_user(self) -> typing.Optional['user.User']:
        """Used by the system to manage authentication behaviors"""
        if self.session and self.session.user:
            await self.session.user.fetch_last_seen_at()
            return self.session.user
        token = self.request.headers.get('Private-Token', None)
        if token:
            current_user = user.User(self.application, token=token)
            if await current_user.authenticate():
                return current_user

    def get_template_namespace(self) -> dict:
        """Returns a dictionary to be used as the default template namespace.

        The results of this method will be combined with additional defaults
        in the :mod:`tornado.template` module and keyword arguments to
        :meth:`~tornado.web.RequestHandler.render`
        or :meth:`~tornado.web.RequestHandler.render_string`.

        """
        namespace = super(RequestHandler, self).get_template_namespace()
        namespace.update({'version': version})
        return namespace

    def on_postgres_error(self, metric_name: str,
                          exc: Exception) -> Exception | None:
        self.logger.exception('processing postgres error: %s',
                              exc.__class__.__name__,
                              exc_info=exc)
        return super().on_postgres_error(metric_name, exc)

    def on_postgres_timing(self, metric_name: str, duration: float) -> None:
        """Invoked by sprockets-postgres after each query"""
        self.application.loop.add_callback(
            self.application.stats.add_duration, {
                'key': 'postgres_query_duration',
                'query': metric_name,
                'endpoint': self.NAME
            }, duration)

    def send_response(self, value: typing.Union[dict, list]) -> None:
        """Send the response to the client"""
        if 'self' not in self._links:
            self._add_self_link(self.request.path)
            self._add_link_header()
        if hasattr(self, 'TTL') and \
                not self.request.headers.get('Pragma') == 'no-cache':
            self._add_response_caching_headers(self.TTL)
        super().send_response(value)

    def set_default_headers(self) -> None:
        """Override the default headers, setting the Server response header"""
        super().set_default_headers()
        self.set_header('Server', self.settings['server_header'])

    def write_error(self, status_code, **kwargs):
        if self._respond_with_html:
            return self.render(
                'error.html',
                javascript_url=self.application.settings.get('javascript_url'),
                status_code=status_code,
                **kwargs)
        super().write_error(status_code, **kwargs)

    def _add_last_modified_header(self, value: datetime.datetime) -> None:
        """Add a RFC-822 formatted timestamp for the Last-Modified HTTP
        response header.

        """
        if not value:
            return
        self.set_header('Last-Modified', self._rfc822_date(value))

    def _add_link_header(self) -> None:
        """Takes the accumulated links and creates a link header value"""
        links = []
        for rel, path in self._links.items():
            links.append('<{}://{}{}>; rel="{}"'.format(
                self.request.protocol, self.request.host, path, rel))
        if links:
            self.add_header('Link', ','.join(links))

    def _add_self_link(self, path: str) -> None:
        """Adds the self Link response header"""
        self._links['self'] = path

    def _add_response_caching_headers(self, ttl: int) -> None:
        """Adds the cache response headers for the object being returned."""
        self.add_header('Cache-Control', 'public, max-age={}'.format(ttl))

    @property
    def _respond_with_html(self) -> bool:
        """Returns True if the current response should respond with HTML"""
        return self.get_response_content_type().startswith(self.TEXT_HTML)

    @staticmethod
    def _rfc822_date(value: datetime.datetime) -> str:
        """Return an RFC-822 formatted timestamp for the given value"""
        return utils.format_datetime(value)


class AuthenticatedRequestHandler(RequestHandler):
    """RequestHandler base class for authenticated requests"""
    async def prepare(self) -> None:
        await super().prepare()

        # Don't require authorization for pre-flight requests
        if self.cors.is_preflight:
            return

        if not self._current_user:
            if self._respond_with_html:
                return await self.render(
                    'index.html',
                    javascript_url=self.application.settings.get(
                        'javascript_url'))
            self.set_status(401)
            await self.finish()
            raise web.Finish()


class ValidatingRequestHandler(AuthenticatedRequestHandler):
    """Validates the request against the OpenAPI spec"""
    async def prepare(self) -> None:
        await super().prepare()

        # Don't validate preflight OPTIONS requests... otherwise,
        # we need to include the OPTIONS stanza in every OpenAPI
        # path spec
        if self.cors.is_preflight:
            return

        try:
            self.application.validate_request(self.request)
        except openapi_core.casting.schemas.exceptions.CastError as err:
            raise errors.InternalServerError(
                'OpenAPI spec casting failure in %s: %s',
                self.__class__.__name__,
                err,
                title='OpenAPI Cast Error')
        except DeserializeError as err:
            raise errors.BadRequest('Failed to deserialize body: %s',
                                    err,
                                    detail=str(err))
        except InvalidSecurity as err:
            raise errors.InternalServerError(
                'Invalid OpenAPI spec security: %s',
                err,
                title='OpenAPI Security Error')
        except OperationNotFound as err:
            raise errors.MethodNotAllowed(err.method.upper())
        except MediaTypeNotFound as err:
            raise errors.UnsupportedMediaType(err.mimetype)
        except PathNotFound as err:
            raise errors.InternalServerError('OpenAPI Spec Error: %s',
                                             err,
                                             title='OpenAPI Spec Error',
                                             detail=str(err))
        except ValidateError as err:
            raise errors.BadRequest(
                'Request failed to validate: %s',
                err,
                detail='The request did not validate',
                errors=[str(e).split('\n')[0] for e in err.schema_errors])


class CRUDRequestHandler(ValidatingRequestHandler):
    """CRUD request handler to reduce large amounts of duplicated code"""

    NAME = 'default'
    DEFAULTS = {}
    ID_KEY: typing.Union[str, list] = 'id'
    IS_COLLECTION = False
    FIELDS = None
    ITEM_NAME = None  # Used to create link headers for POST requests
    TTL = 300

    OMIT_FIELDS: typing.Optional[list] = None  # currently unused
    """Set this to omit specific fields from a record response.

    Set this to `None` for the previous behavior of omitting
    the `created_at` and `last_modified_at` fields.

    Set this to the empty list to pass *all* fields in the
    response.

    Set this to a list of names to explicitly exclude specific
    fields in the response.

    """

    DELETE_SQL: typing.Optional[str] = None
    GET_SQL: typing.Optional[str] = None
    PATCH_SQL: typing.Optional[str] = None
    POST_SQL: typing.Optional[str] = None

    def initialize(self) -> None:
        super().initialize()
        if self.__class__.OMIT_FIELDS is None and self.OMIT_FIELDS is None:
            # implement backwards compatible behavior
            self.OMIT_FIELDS = ['created_at', 'last_modified_at']

    async def delete(self, *args, **kwargs):
        if self.DELETE_SQL is None:
            self.logger.debug('DELETE_SQL not defined')
            raise errors.MethodNotAllowed('DELETE')
        await self._delete(kwargs)

    async def get(self, *args, **kwargs):
        if self.GET_SQL is None:
            self.logger.debug('GET_SQL not defined')
            raise errors.MethodNotAllowed('GET')
        if self._respond_with_html:
            return self.render(
                'index.html',
                javascript_url=self.application.settings.get('javascript_url'))
        await self._get(kwargs)

    async def patch(self, *args, **kwargs):
        if self.PATCH_SQL is None:
            self.logger.debug('PATCH_SQL not defined')
            raise errors.MethodNotAllowed('PATCH')
        await self._patch(kwargs)

    async def post(self, *args, **kwargs):
        if self.POST_SQL is None:
            self.logger.debug('POST_SQL not defined')
            raise errors.MethodNotAllowed('POST')
        await self._post(kwargs)

    def send_response(self, value: typing.Union[dict, list]) -> None:
        """Send the response to the client"""
        if isinstance(value, list):
            return super().send_response(value)

        if self.request.method != 'GET' or not self.IS_COLLECTION:
            self._add_last_modified_header(
                value.get('last_modified_at', value.get('created_at')))
            if self.ID_KEY:
                if isinstance(self.ID_KEY, list):
                    args = [str(value[k]) for k in self.ID_KEY]
                else:
                    args = [str(value[self.ID_KEY])]

                try:
                    self._add_self_link(
                        self.reverse_url(self.ITEM_NAME or self.NAME, *args))
                except (AssertionError, KeyError):
                    self.logger.debug('Failed to reverse URL for %s %r',
                                      self.NAME, args)
            self._add_link_header()

        for name in self.OMIT_FIELDS:
            value.pop(name, None)

        super().send_response(value)

    async def _delete(self, kwargs):
        result = await self.postgres_execute(self.DELETE_SQL,
                                             self._get_query_kwargs(kwargs),
                                             'delete-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)
        self.set_status(204, reason='Item Deleted')

    async def _get(self, kwargs):
        result = await self.postgres_execute(self.GET_SQL,
                                             self._get_query_kwargs(kwargs),
                                             'get-{}'.format(self.NAME))
        if not result.row_count or not result.row:
            raise errors.ItemNotFound(instance=self.request.uri)
        for key, value in result.row.items():
            if isinstance(value, uuid.UUID):
                result.row[key] = str(value)
        self.send_response(result.row)

    def _get_query_kwargs(self, kwargs) -> dict:
        if isinstance(self.ID_KEY, list):
            return {k: kwargs[k] for k in self.ID_KEY}
        return {self.ID_KEY: kwargs[self.ID_KEY]}

    async def _patch(self, kwargs):
        patch_value = self.get_request_body()

        result = await self.postgres_execute(self.GET_SQL,
                                             self._get_query_kwargs(kwargs),
                                             'get-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.ItemNotFound(instance=self.request.uri)

        original = dict(result.row)
        for key in {
                'created_at', 'created_by', 'last_modified_at',
                'last_modified_by', 'recorded_at', 'recorded_by'
        }:
            if key in original:
                del original[key]

        for key, value in original.items():
            if isinstance(value, uuid.UUID):
                original[key] = str(value)

        # Apply the patch to the current value
        try:
            patch = jsonpatch.JsonPatch(patch_value)
            updated = patch.apply(original)
        except (jsonpatch.JsonPatchException,
                jsonpointer.JsonPointerException) as error:
            raise errors.UnprocessableEntity('Failed to apply JSON patch: %s',
                                             error,
                                             detail=str(error))

        # Bail early if there are no changes
        if not {
                k: original[k]
                for k in original if k in updated and original[k] != updated[k]
        }:
            self._add_self_link(self.request.path)
            self._add_link_header()
            return self.set_status(304)

        # Let the endpoint have a say as to the validity
        if not self._check_validity(updated):
            self.logger.debug('Invalid representation generated %r', updated)
            raise errors.BadRequest('Invalid instance generated by update')

        if isinstance(self.ID_KEY, list):
            for key in self.ID_KEY:
                updated['current_{}'.format(key)] = kwargs[key]
        else:
            updated['current_{}'.format(self.ID_KEY)] = kwargs[self.ID_KEY]
        updated['username'] = self._current_user.username

        result = await self.postgres_execute(self.PATCH_SQL, updated,
                                             'patch-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.DatabaseError('No rows returned from PATCH_SQL',
                                       title='Failed to update record')

        # Send the new record as a response
        await self._get(self._get_query_kwargs(updated))

    async def _post(self, kwargs, overrides=None) -> dict:
        values = self.get_request_body().copy()  # avoid mutating cached copy

        # Handle compound keys for child object CRUD
        if isinstance(self.ID_KEY, list):
            for key in self.ID_KEY:
                if key not in values and key in kwargs:
                    values[key] = kwargs[key]
        elif self.ID_KEY not in values and self.ID_KEY in kwargs:
            values[self.ID_KEY] = kwargs[self.ID_KEY]

        # Set defaults of None for all fields in insert
        for name in self.FIELDS:
            if name not in values:
                values[name] = self.DEFAULTS.get(name)

        values['username'] = self._current_user.username
        if overrides:
            values.update(overrides)

        result = await self.postgres_execute(self.POST_SQL, values,
                                             'post-{}'.format(self.NAME))
        if not result.row_count:
            raise errors.DatabaseError('No rows returned from POST_SQL',
                                       title='Failed to create record')

            # Return the record as if it were a GET
        await self._get(self._get_query_kwargs(result.row))
        return result.row

    @staticmethod
    def _check_validity(instance: dict[str, typing.Any]) -> bool:
        """Override this method to check if an instance after patching"""
        return True


class CollectionRequestHandler(CRUDRequestHandler):
    DEFAULTS = {}
    ID_KEY: typing.Union[str, list] = 'id'
    IS_COLLECTION: True
    FIELDS = None
    ITEM_NAME = None  # Used to create link headers for POST requests
    COLLECTION_SQL = """SELECT * FROM pg_tables WHERE schemaname = 'v1';"""
    TTL = 300

    async def get(self, *args, **kwargs):
        result = await self.postgres_execute(self.COLLECTION_SQL,
                                             kwargs,
                                             metric_name='get-{}'.format(
                                                 self.NAME))
        self.send_response(result.rows)


class AdminCRUDRequestHandler(CRUDRequestHandler):
    @require_permission('admin')
    async def delete(self, *args, **kwargs):
        await super().delete(*args, **kwargs)

    @require_permission('admin')
    async def get(self, *args, **kwargs):
        await super().get(*args, **kwargs)

    @require_permission('admin')
    async def patch(self, *args, **kwargs):
        await super().patch(*args, **kwargs)

    @require_permission('admin')
    async def post(self, *args, **kwargs):
        await super().post(*args, **kwargs)


@dataclasses.dataclass
class TimeBasedPaginationToken:
    """Opaque token used in "next" links by the PaginatedRequestMixin"""

    start: datetime.datetime
    limit: int
    earliest: datetime.datetime

    def with_start(self, start: datetime.datetime) -> TimeBasedPaginationToken:
        """Create a new token with a different start value"""
        return TimeBasedPaginationToken(start=start,
                                        limit=self.limit,
                                        earliest=self.earliest)

    @classmethod
    def from_header(cls, header_val: str) -> TimeBasedPaginationToken:
        """Create a token from an opaque header value"""
        raw_token = common.urlsafe_padded_b64decode(header_val)
        data = umsgpack.unpackb(raw_token)
        return TimeBasedPaginationToken(
            earliest=datetime.datetime.fromisoformat(data['earliest']),
            limit=data['limit'],
            start=(datetime.datetime.fromisoformat(data['start']) -
                   datetime.timedelta.resolution))

    def to_header(self) -> str:
        """Generate an opaque header value from this token"""
        return base64.urlsafe_b64encode(
            umsgpack.packb({
                'earliest': self.earliest.isoformat(),
                'limit': self.limit,
                'start': self.start.isoformat(),
            })).decode('ascii').rstrip('=')


TimeBasedPaginationQueryMethod = typing.Callable[
    [dict[str, typing.Any]],
    typing.Coroutine[None, None, list[tuple[datetime.datetime, typing.Any]]]]
"""Type signature of the function called by PaginatedRequestHandler

This is something that looks like:

  async def retrieve_data(
      params: dict[str, typing.Any]) -> list[tuple[datetime.datetime, dict]]:
    ...

"""


class TimeBasedPaginationMixin(web.RequestHandler):
    """Mix this in to repeatedly retrieve query results ordered by time

    Call the `fetch_items` method to call a specified query method
    repeatedly until "token.limit" items have been retrieved.  The
    query method is called with a well-known set of parameters and
    returns a list of ("when", "data") tuples.  The query parameters
    are passed as a dictionary that contains at least the following
    keys:

    * ``earlier``   the earliest (oldest) record to query for
    * ``later``     the most recent record to query for
    * ``remaining`` the maximum number of records to return

    The query function should build a query similar that contains
    ``... some_column BETWEEN %(earlier)s AND %(later)s ...`` as
    well as ``... LIMIT %(remaining)s``.  The parameters dictionary
    contains any additional parameters that are passed into the
    ``fetch_items`` call.

    Call the `pagination_token_from_request` method to retrieve the
    pagination token from the current request if one exists.  If
    the token does not exist, then the concrete implementation needs
    to generate one to pass into `fetch_items`.

    """

    logger: logging.Logger

    async def fetch_items(self, token: TimeBasedPaginationToken,
                          query: TimeBasedPaginationQueryMethod,
                          time_step: datetime.timedelta,
                          **initial_params) -> list[dict]:
        """Call this to iterate over a query function and collect the result

        :param token: controls the pagination process
        :param query: coroutine function to call to retrieve data
        :param time_step: amount of time to "step" for each iteration.
            This controls how much data the "query" function is asked
            to retrieve.
        :param initial_params: additional parameters to include when
            calling `query`
        :returns: list of calling `query` until `token.limit` items
            are retrieved

        """
        buckets = []
        params = initial_params.copy()
        params.update({
            'earlier': token.start - time_step,
            'later': token.start,
            'remaining': token.limit,
        })
        while params['remaining'] > 0 and params['later'] >= token.earliest:
            buckets.extend(await query(params))
            buckets.sort(key=operator.itemgetter(0), reverse=True)
            params.update({
                'earlier': params['earlier'] - time_step,
                'later': params['earlier'],
                'remaining': token.limit - len(buckets),
            })

        buckets = buckets[:token.limit]
        if buckets and buckets[-1][0] > token.earliest:
            target = yarl.URL(self.request.uri).with_query(
                token=token.with_start(buckets[-1][0]).to_header())
            header = datastructures.LinkHeader(str(target), [('rel', 'next')])
            self.add_header('Link', str(header))

        return [entry for _, entry in buckets]

    def get_pagination_token_from_request(
            self) -> TimeBasedPaginationToken | None:
        """Retrieve the pagination token from the request"""
        token = self.get_query_argument('token', '')
        if token:
            try:
                return TimeBasedPaginationToken.from_header(token)
            except ValueError as error:
                self.logger.warning('ignoring malformed token %r: %s', token,
                                    error)


class PaginationToken:
    """Opaque pagination token for use with PaginatedCollectionHandler

    Pagination tokens are passed as the `token` query parameter
    and are read from the request by the `from_request` class method.
    The only field that is in the default token is `limit` which
    defaults to 100.

    Be aware that the body of the token is used as parameters to
    the pagination query in the handler. This interplay between
    the two classes is what should form the basis of your token.

    Create a new subclass that contains additional content to
    store in the token. There are a few rules that need to be
    followed:

    1. the keyword parameter names in `__init__` are required
       to match the keys in the dict returned from `as_dict`
    2. the `limit` parameter cannot be removed
    3. all parameters to the initializer must have default
       values unless they are path parameters to the handler
    4. path parameter names must have corresponding initializer
       parameter names

    The easiest way to do this is to set the values that you
    want for the first page as the defaults for the initializer
    keyword params and add matching path parameter names. If the
    token is not present in the request, the default token is
    created by calling the initializer without only the path
    parameters.

    Then implement the `with_first` method to merge attributes
    from the value parameter into `self.as_dict()` by passing
    overrides.

    .. code-block::

       class MyToken(PaginationToken):
          def __init__(self, *, name: str = '', project_id: int | str,
                       **kwargs: object) -> None:
             super().__init__(name=name, project_id=project_id, **kwargs)
          def with_first(self, value: dict[str, object]) -> typing.Self:
             kwargs = self.as_dict(name=value['name'])
             return MyToken(**kwargs)

    """
    def __init__(self, *, limit: int = 100, **other_props: object):
        self._data: dict[str, object] = {'limit': limit}
        self._data.update(other_props)

    @classmethod
    def from_request(cls, req: httputil.HTTPServerRequest,
                     **path_args: str) -> typing.Self:
        """Generate a token for a request

        If there is no token in the request, then a new token
        is created by calling the initializer without parameters.
        Otherwise, the content of the token is decoded and passed
        as keyword parameters to create the token instance.
        """
        encoded_token = req.query_arguments.get('token')
        if encoded_token:
            try:
                raw_token = common.urlsafe_padded_b64decode(
                    encoded_token[0].decode())
                data = umsgpack.unpackb(raw_token)
            except (AttributeError, ValueError):
                pass
            else:
                # Should we verify that path_args are present
                # in the decoded token?  What happens if someone
                # uses a token on a different URL?
                return cls(**data)

        token = cls(**path_args)
        with contextlib.suppress(IndexError, KeyError):
            token._data['limit'] = int(req.query_arguments['limit'][0])
        return token

    def as_dict(self, **overrides: object) -> dict[str, object]:
        """The properties of the token as a dictionary

        This is used to encode the token into a query string
        value. The `overrides` parameter is a convenience for
        implementing the `with_first` method.
        """
        data = self._data.copy()
        data.update(overrides)
        return data

    def with_first(self, _value: dict[str, object]) -> typing.Self:
        """Return a new token that uses `value` as the search target"""
        raise NotImplementedError

    @property
    def limit(self) -> int:
        return typing.cast(int, self._data['limit'])

    def __str__(self) -> str:
        data = umsgpack.packb(self.as_dict())
        return base64.urlsafe_b64encode(data).decode().rstrip('=')


class PaginatedCollectionHandler(CollectionRequestHandler):
    """Implements a paginated `get` method

    The `get` method uses a combination of the token returned
    from `get_pagination_token_from_request` and the COLLECTION_SQL
    query to retrieve and return pages of the collection.

    Note that the dict form of the token is passed as the parameters
    to the COLLECTION_SQL query.

    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.ID_KEY is None:
            self._extract_keys = lambda itm: None
        elif isinstance(self.ID_KEY, str):
            self._extract_keys = lambda itm: (itm[self.ID_KEY], )
        else:
            self._extract_keys = lambda itm: tuple(itm[k] for k in self.ID_KEY)

    def get_pagination_token_from_request(
            self, **path_params: str) -> PaginationToken:
        """Return a pagination token of the appropriate type

        The keyword parameters will match the named patterns
        from the URL match. These correspond with the keyword
        parameters passed to the `get` method.
        """
        raise NotImplementedError()

    async def get(self, **kwargs: str) -> None:
        token = self.get_pagination_token_from_request(**kwargs)
        params = token.as_dict()
        params['limit'] += 1
        rsp = await self.postgres_execute(self.COLLECTION_SQL, params)
        items = rsp.rows
        url = yarl.URL(self.request.uri)
        if len(items) > token.limit:
            items = items[:token.limit]
            target = url.with_query(token=str(token.with_first(items[-1])))
            self.add_header('Link', common.build_link_header(target, 'next'))
        target = url.with_query(limit=token.limit)
        self.add_header('Link', common.build_link_header(target, 'first'))

        if self.ITEM_NAME is not None:
            for item in items:
                item.setdefault(
                    'link',
                    self.reverse_url(self.ITEM_NAME,
                                     *self._extract_keys(item)),
                )
        if hasattr(self, '_postprocess_item'):
            for item in items:
                self._postprocess_item(item)

        self.send_response(items)


ModelType = typing.TypeVar('ModelType', bound=pydantic.BaseModel)


class PydanticHandlerMixin(RequestHandler):
    def parse_request_body_as(
            self,
            model_cls: typing.Type[ModelType],
            context: typing.Optional[dict] = None) -> ModelType:
        try:
            return model_cls.model_validate(self.get_request_body(),
                                            context=context)
        except pydantic.ValidationError as error:
            raise errors.UnprocessableEntity(
                'Failed to validate request body: %s',
                error,
                validation_errors=error.errors(include_input=False,
                                               include_url=False))



import http.client
import unittest.mock
import uuid

import sprockets.mixins.http
from tornado import httputil

from imbi import errors, models, version
from imbi.clients import pagerduty
from tests import base


class PagerDutyClientTestCase(base.TestCaseWithReset):
    ADMIN_ACCESS = True
    TRUNCATE_TABLES = ['v1.integrations']

    def setUp(self) -> None:
        super().setUp()
        self._patcher = unittest.mock.patch.object(
            pagerduty._PagerDutyClient,
            'http_fetch',
            new_callable=unittest.mock.AsyncMock,
            create=True,
        )
        self.http_fetch_mock = self._patcher.start()
        self.integration = self.create_integration(
            api_endpoint='https://example.com/pagerduty/api',
            api_secret='my-secret')

        config = self.settings.setdefault('automations', {})
        config['pagerduty'] = {'enabled': True}

        self.client: pagerduty._PagerDutyClient = self.run_until_complete(
            pagerduty.create_client(self.app, self.integration['name']))

    def tearDown(self) -> None:
        self._patcher.stop()
        super().tearDown()

    @staticmethod
    def create_successful_response(body=None):
        if body is None:
            body = {
                'service': {
                    'html_url': 'https://example.com/html_url',
                    'id': str(uuid.uuid4()),
                    'self': 'https://example.com/self',
                }
            }
        response = unittest.mock.Mock()
        response.ok = True
        response.code = http.HTTPStatus.CREATED.value
        response.body = body
        return response

    @staticmethod
    def create_error_response() -> sprockets.mixins.http.HTTPResponse:
        response = unittest.mock.Mock()
        response.ok = False
        response.code = http.HTTPStatus.INTERNAL_SERVER_ERROR.value
        response.history = [unittest.mock.Mock()]
        response.history[0].effective_url = 'https://example.com/effective-url'
        return response


class PagerDutyCreateClientTests(base.TestCaseWithReset):
    ADMIN_ACCESS = True
    TRUNCATE_TABLES = ['v1.integrations', 'v1.namespaces']

    def setUp(self) -> None:
        super().setUp()
        self.integration = self.create_integration(api_secret='my-secret')
        config = self.settings.setdefault('automations', {})
        config['pagerduty'] = {'enabled': True}

    def test_correctly_configured(self) -> None:
        self.run_until_complete(
            pagerduty.create_client(self.app, self.integration['name']))

    def test_missing_configuration(self) -> None:
        del self.settings['automations']['pagerduty']
        with self.assertRaises(errors.ClientUnavailableError):
            self.run_until_complete(
                pagerduty.create_client(self.app, self.integration['name']))

    def test_pagerduty_disabled(self) -> None:
        self.settings['automations']['pagerduty']['enabled'] = False
        with self.assertRaises(errors.ClientUnavailableError):
            self.run_until_complete(
                pagerduty.create_client(self.app, self.integration['name']))

    def test_with_incorrect_integration_name(self) -> None:
        with self.assertRaises(errors.ClientUnavailableError):
            self.run_until_complete(
                pagerduty.create_client(self.app, str(uuid.uuid4())))

    def test_with_missing_api_secret(self) -> None:
        integration = self.create_integration()
        with self.assertRaises(errors.ClientUnavailableError):
            self.run_until_complete(
                pagerduty.create_client(self.app, integration['name']))


class PagerDutyProjectCreationTests(PagerDutyClientTestCase):
    TRUNCATE_TABLES = [
        'v1.environments', 'v1.integrations', 'v1.namespaces', 'v1.projects',
        'v1.project_types'
    ]

    def setUp(self) -> None:
        super().setUp()
        project = self.create_project()
        self.disabled_project: models.Project = self.run_until_complete(
            models.project(project['id'], self.app))

        enabled_namespace = self.create_namespace(pagerduty_policy='my-policy')
        project = self.create_project(namespace_id=enabled_namespace['id'])
        self.enabled_project: models.Project = self.run_until_complete(
            models.project(project['id'], self.app))

    def test_calling_with_namespace_policy(self) -> None:
        self.http_fetch_mock.return_value = self.create_successful_response()
        self.run_until_complete(
            self.client.create_service(self.enabled_project))
        self.http_fetch_mock.assert_awaited_once_with(
            'https://example.com/pagerduty/api/services',
            method='POST',
            body={
                'name': self.enabled_project.slug,
                'description': self.enabled_project.description,
                'escalation_policy': {
                    'id': self.enabled_project.namespace.pagerduty_policy,
                    'type': 'escalation_policy_reference'
                }
            },
            request_headers=unittest.mock.ANY,
            content_type='application/json',
            user_agent=f'imbi/{version} (PagerDutyClient)')
        _, kwargs = self.http_fetch_mock.await_args
        request_headers = httputil.HTTPHeaders(kwargs['request_headers'])
        self.assertEqual('Token token=my-secret',
                         request_headers['Authorization'])

    def test_calling_without_namespace_policy(self) -> None:
        with self.assertRaises(errors.InternalServerError):
            self.run_until_complete(
                self.client.create_service(self.disabled_project))
        self.http_fetch_mock.assert_not_awaited()

    def test_empty_description_is_omitted(self) -> None:
        self.http_fetch_mock.return_value = self.create_successful_response()
        self.enabled_project.description = None
        self.run_until_complete(
            self.client.create_service(self.enabled_project))
        self.http_fetch_mock.assert_awaited_once_with(
            'https://example.com/pagerduty/api/services',
            method='POST',
            body={
                'name': self.enabled_project.slug,
                'escalation_policy': {
                    'id': self.enabled_project.namespace.pagerduty_policy,
                    'type': 'escalation_policy_reference'
                }
            },
            request_headers=unittest.mock.ANY,
            content_type='application/json',
            user_agent=f'imbi/{version} (PagerDutyClient)')

    def test_unexpected_response_handling(self) -> None:
        self.http_fetch_mock.return_value = self.create_successful_response({})
        with self.assertRaises(errors.InternalServerError):
            self.run_until_complete(
                self.client.create_service(self.enabled_project))

    def test_pagerduty_api_failure(self) -> None:
        response = unittest.mock.Mock()
        response.ok = False
        response.code = http.HTTPStatus.INTERNAL_SERVER_ERROR.value
        response.history = [
            unittest.mock.Mock(effective_url='https://example.com')
        ]
        self.http_fetch_mock.return_value = response

        with self.assertRaises(errors.InternalServerError):
            self.run_until_complete(
                self.client.create_service(self.enabled_project))


class PagerDutyProjectDeletionTests(PagerDutyClientTestCase):
    def test_removing_service(self) -> None:
        self.run_until_complete(self.client.remove_service('some-service-id'))
        self.http_fetch_mock.assert_awaited_once_with(
            'https://example.com/pagerduty/api/services/some-service-id',
            method='DELETE',
            request_headers=unittest.mock.ANY,
            user_agent=f'imbi/{version} (PagerDutyClient)')
        _, kwargs = self.http_fetch_mock.await_args
        request_headers = httputil.HTTPHeaders(kwargs['request_headers'])
        self.assertEqual('Token token=my-secret',
                         request_headers['Authorization'])


class PagerDutyCreateIntegrationTests(PagerDutyClientTestCase):
    TRUNCATE_TABLES = [
        'v1.environments', 'v1.integrations', 'v1.namespaces', 'v1.projects',
        'v1.project_secrets', 'v1.project_types'
    ]

    def setUp(self) -> None:
        super().setUp()
        self.namespace = self.create_namespace(pagerduty_policy='my-policy')
        project = self.create_project()
        self.project: models.Project = self.run_until_complete(
            models.project(project['id'], self.app))
        self.service = pagerduty.ServiceInfo(
            id='SERVICE',
            html_url='https://example.com/html_url',
            self='https://example.com/self')

    def test_creating_integration(self) -> None:
        self.http_fetch_mock.return_value = self.create_successful_response({
            'integration': {
                'id': 'INTEGRATION',
                'name': 'my-hook',
                'type': 'whatever',
                'self': 'https://example.com/self',
                'html_url': 'https://example.com/html_url',
                'integration_key': 'super secret key',
            }
        })
        integration = self.run_until_complete(
            self.client.create_inbound_api_integration('my-hook',
                                                       self.service))
        self.http_fetch_mock.assert_awaited_once_with(
            'https://example.com/pagerduty/api/services/SERVICE/integrations',
            method='POST',
            body={
                'integration': {
                    'name': 'my-hook',
                    'type': 'generic_events_api_inbound_integration',
                }
            },
            content_type='application/json',
            request_headers=unittest.mock.ANY,
            user_agent=f'imbi/{version} (PagerDutyClient)',
        )
        self.assertEqual('INTEGRATION', integration.id)
        self.assertEqual('super secret key', integration.integration_key)

    def test_api_failure(self) -> None:
        self.http_fetch_mock.return_value = self.create_error_response()
        with self.assertRaises(errors.InternalServerError):
            self.run_until_complete(
                self.client.create_inbound_api_integration(
                    'my-hook', self.service))



import re

from imbi.endpoints import base


class RequestHandler(base.AuthenticatedRequestHandler):
    NAME = 'reports-component-usage'
    SQL = re.sub(
        r'\s+', ' ', """\
        SELECT c.name, c.package_url, c.active_version, c.status,
               COUNT(v.id) AS version_count,
               COUNT(p.project_id) AS project_count
          FROM v1.components AS c
          JOIN v1.component_versions AS v ON v.package_url = c.package_url
          LEFT JOIN v1.project_components AS p ON p.version_id = v.id
         GROUP BY c.name, c.package_url, c.active_version
        """)

    async def get(self) -> None:
        result = await self.postgres_execute(self.SQL, metric_name=self.NAME)
        self.send_response(result.rows)



from __future__ import annotations

import asyncio
import datetime
import re
import typing

from tornado import web

from imbi.endpoints import base


class CollectionRequestHandler(base.TimeBasedPaginationMixin,
                               base.ValidatingRequestHandler):

    NAME = 'activity-feed'

    PROJECT_FEED_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT 'ProjectFeedEntry' AS "type", "when",
               namespace_id, namespace, project_id, project_name,
               project_type, who, display_name, email_address,
               what
          FROM v1.activity_feed
         WHERE "when" >  %(earlier)s
           AND "when" <= %(later)s
         ORDER BY "when" DESC
         LIMIT %(remaining)s
        """)

    OPERATIONS_LOG_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT o.id, o.recorded_at, o.recorded_by, o.completed_at,
               o.project_id, o.environment, o.change_type, o.description,
               o.link, o.notes, o.ticket_slug, o.version,
               COALESCE(
                  u.email_address,
                  COALESCE(u2.email_address, 'UNKNOWN')
               ) AS email_address,
               COALESCE(
                  u.display_name,
                  COALESCE(u2.display_name, o.recorded_by)
               ) AS display_name,
               p.name AS project_name,
               'OperationsLogEntry' AS "type",
               o.occurred_at,
               o.performed_by
          FROM v1.operations_log AS o
          LEFT JOIN v1.projects AS p ON p.id = o.project_id
          LEFT JOIN v1.users AS u ON u.username = o.performed_by
          LEFT JOIN v1.users AS u2 ON u2.username = o.recorded_by
         WHERE recorded_at >  %(earlier)s
           AND recorded_at <= %(later)s
         ORDER BY o.recorded_at DESC, o.id DESC
         LIMIT %(remaining)s
        """)

    EARLIEST_EVENT_SQL = re.sub(
        r'\s+', ' ', """\
        WITH T AS (
            SELECT MIN("when") AS earliest FROM v1.activity_feed
            UNION
            SELECT MIN(recorded_at) AS earliest FROM v1.operations_log
        )
        SELECT COALESCE(MIN(earliest), CURRENT_TIMESTAMP) AS earliest
          FROM T
        """)

    @web.authenticated
    async def get(self):
        token = await self._get_pagination_token()
        buckets = await self.fetch_items(
            token,
            self._retrieve_rows,
            datetime.timedelta(days=10),
            omit_users=set(self.get_query_arguments('omit_user')))

        self.send_response(buckets)

    async def _retrieve_rows(
            self, params) -> list[tuple[datetime.datetime, typing.Any]]:
        activity_feed, ops_log = await asyncio.gather(
            self.postgres_execute(self.PROJECT_FEED_SQL, params),
            self.postgres_execute(self.OPERATIONS_LOG_SQL, params))
        self.logger.debug('fetched page containing %s activity, %s ops log',
                          len(activity_feed), len(ops_log))

        all_rows = [(row['when'], row) for row in activity_feed
                    if row['who'] not in params['omit_users']]
        all_rows.extend((row['recorded_at'], row) for row in ops_log
                        if row['recorded_by'] not in params['omit_users'])

        return all_rows

    async def _get_pagination_token(self):
        token = self.get_pagination_token_from_request()
        if token is None:
            self.logger.info('creating new token')
            result = await self.postgres_execute(self.EARLIEST_EVENT_SQL)
            token = base.TimeBasedPaginationToken(
                start=datetime.datetime.now(datetime.timezone.utc),
                limit=int(self.get_query_argument('limit', '25')),
                earliest=result.row['earliest'],
            )
        return token



from psycopg2 import sql

from imbi.endpoints import base


class UserRequestHandler(base.AuthenticatedRequestHandler):

    NAME = 'ui-user'

    def get(self, *args, **kwargs):
        user = self.current_user.as_dict()
        del user['password']
        self.send_response(user)


class AvailableAutomationsHandler(base.AuthenticatedRequestHandler):

    NAME = 'available-automations'

    async def get(self, *args, **kwargs):
        automation_categories = self.get_query_arguments('category')
        project_type_ids = self.get_query_arguments('project_type_id')
        project_types = self.get_query_arguments('project_type')

        query = sql.SQL('SELECT DISTINCT a.name AS automation_name,'
                        '                a.categories,'
                        '                i.name AS integration_name,'
                        '                a.slug AS automation_slug'
                        '          FROM v1.integrations AS i'
                        '          JOIN v1.automations AS a'
                        '            ON a.integration_name = i.name'
                        '     LEFT JOIN v1.user_oauth2_tokens AS t'
                        '            ON t.integration = i.name'
                        '         WHERE (i.api_secret IS NOT NULL'
                        '                OR t.username = %(username)s)')
        filter_clauses = []
        identification_clauses = []
        params = {'username': self._current_user.username}

        if automation_categories:
            filter_clauses.append(
                sql.SQL('a.categories && ARRAY[%(categories)s]::{}[]').format(
                    sql.Identifier('v1', 'automation_category_type')))
            params['categories'] = automation_categories

        if project_type_ids:
            identification_clauses.append(
                sql.SQL('EXISTS (SELECT aa.automation_id'
                        '          FROM v1.available_automations AS aa'
                        '         WHERE aa.automation_id = a.id'
                        '           AND aa.project_type_id IN %(type_ids)s)'))
            params['type_ids'] = tuple(project_type_ids)
        if project_types:
            identification_clauses.append(
                sql.SQL('EXISTS (SELECT aa.automation_id'
                        '          FROM v1.available_automations AS aa'
                        '          JOIN v1.project_types AS pt'
                        '            ON pt.id = aa.project_type_id'
                        '         WHERE aa.automation_id = a.id'
                        '           AND pt.slug IN %(project_types)s)'))
            params['project_types'] = tuple(project_types)

        if filter_clauses:
            query = sql.SQL(' AND ').join([query] + filter_clauses)
        if identification_clauses:
            clause = sql.SQL('(') + sql.SQL(' OR ').join(
                identification_clauses) + sql.SQL(')')
            query = sql.SQL(' AND ').join([query, clause])

        result = await self.postgres_execute(query, params)

        # arrays of custom types come out as strings like '{foo,bar,baz}'
        response = [{
            **row,
            'categories': row['categories'].lstrip('{').rstrip('}').split(',')
        } for row in result.rows]
        self.send_response(response)



from __future__ import annotations

import logging
import typing

import pydantic
import sprockets.mixins.http
import yarl

from imbi import errors, models, version
if typing.TYPE_CHECKING:
    from imbi import app


class _ProjectCreationResponse(pydantic.BaseModel):
    name: str
    slug: str
    id: str


class _ProjectKeysDsn(pydantic.BaseModel):
    public: pydantic.HttpUrl
    secret: pydantic.HttpUrl


class _ProjectKeysResponse(pydantic.BaseModel):
    dsn: _ProjectKeysDsn
    id: str
    isActive: bool
    projectId: int


class ProjectInfo(pydantic.BaseModel):
    name: str
    slug: str
    id: str
    link: pydantic.HttpUrl
    keys: dict[str, str]


async def create_client(
    application: app.Application,
    integration_name: str,
) -> '_SentryClient':
    """Create a sentry API client

    The settings needed to create a client are in the application
    settings and the database currently. This function finds them
    and creates a client.  If the configuration is invalid or the
    integration is disabled, a ``ClientUnavailableError`` is raised
    with an appropriate message.
    """
    logger = logging.getLogger(__package__).getChild('create_client')
    settings = application.settings['automations']['sentry']
    if not settings['enabled']:
        raise errors.ClientUnavailableError(integration_name, 'disabled')
    if not settings.get('organization'):
        raise errors.ClientUnavailableError(integration_name,
                                            'organization is not configured')

    sentry_info = await models.integration(integration_name, application)
    if not sentry_info:
        logger.warning('%r integration is enabled by not configured',
                       integration_name)
        raise errors.ClientUnavailableError(integration_name, 'not configured')

    return _SentryClient(yarl.URL(str(sentry_info.api_endpoint)),
                         sentry_info.api_secret, settings['organization'])


class _SentryClient(sprockets.mixins.http.HTTPClientMixin):
    def __init__(self, api_endpoint: yarl.URL, api_secret: str,
                 organization: str) -> None:
        super().__init__()
        self.logger = logging.getLogger(__package__).getChild('SentryClient')

        self.api_url = api_endpoint
        self.url = api_endpoint.with_path('/')
        self.organization = organization
        self.api_secret = api_secret
        self.headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.api_secret}',
        }

    async def api(self, url: yarl.URL | str, *, method: str = 'GET', **kwargs):
        if not isinstance(url, yarl.URL):
            url = yarl.URL(url)
        if not url.is_absolute():
            url = self.api_url / url.path.lstrip('/')

        # the sentry API wants a trailing slash
        url = str(url).rstrip('/') + '/'

        request_headers = kwargs.pop('request_headers', {})
        request_headers.update(self.headers)
        if kwargs.get('body') is not None:
            request_headers['Content-Type'] = 'application/json'
            kwargs['content_type'] = 'application/json'
        kwargs['request_headers'] = request_headers
        kwargs['user_agent'] = f'imbi/{version} (SentryClient)'
        rsp = await super().http_fetch(str(url), method=method, **kwargs)
        if not rsp.ok:
            self.logger.error('%s %s failed: %s', method, url, rsp.code)
            self.logger.debug('auth token: %s...%s', self.api_secret[:4],
                              self.api_secret[-4:])
            self.logger.debug('response body: %r', rsp.body)
        return rsp

    async def create_project(self, team_slug: str,
                             project_name: str) -> ProjectInfo:
        url = yarl.URL('/teams') / self.organization / team_slug / 'projects'
        response = await self.api(url,
                                  method='POST',
                                  body={'name': project_name})
        if not response.ok:
            self.logger.error('Sentry API failure: body %r', response.body)
            raise errors.InternalServerError(
                'POST %s failed: %s',
                response.history[-1].effective_url,
                response.code,
                title='Sentry API Failure')
        else:
            try:
                data = _ProjectCreationResponse.parse_obj(response.body)
            except pydantic.ValidationError as error:
                self.logger.error('failed to parse response body: %r',
                                  response.body)
                raise errors.InternalServerError(
                    'Failed to parse sentry create_project response: %s',
                    error,
                    title='Sentry Client Failure')

        project = ProjectInfo(
            id=data.id,
            name=data.name,
            slug=data.slug,
            link=str(self.url / 'organizations' / self.organization /
                     'projects' / data.slug),
            keys={})

        # TODO update the project platform based on the project details
        #      so that python projects use "platform=python" and
        #      JS projects use "platform=javascript"
        # Not sure if this is actually necessary so I'm omitting it.
        # https://github.com/AWeber-Imbi/imbi/issues/73
        # ref: https://docs.sentry.io/api/projects/update-a-project/
        # ref: https://docs.sentry.io/platforms/

        url = yarl.URL('/projects') / self.organization / project.slug

        try:
            response = await self.api(url / 'keys')
            if not response.ok:
                raise errors.InternalServerError(
                    'Failed to retrieve sentry project keys: %s',
                    response.code,
                    title='Sentry API Error')
            try:
                data = _ProjectKeysResponse.parse_obj(response.body[0])
            except pydantic.ValidationError as error:
                self.logger.error('failed to parse response body: %r',
                                  response.body)
                raise errors.InternalServerError(
                    'Failed to parse sentry project keys response: %s',
                    error,
                    title='Sentry Client Failure')
            else:
                project.keys.update({
                    'public': data.dsn.public,
                    'secret': data.dsn.secret,
                })

        # If we fail to fetch or extract the keys that the client
        # needs, then remove the project since we cannot use it.
        except errors.ApplicationError as error:
            self.logger.error('removing Sentry project %s due to error: %s',
                              project.slug, error)
            await self.remove_project(project.slug)
            raise

        return project

    async def remove_project(self, project_slug: str) -> None:
        await self.api(yarl.URL('/projects') / self.organization /
                       project_slug,
                       method='DELETE')



"""
Handy Timestamp Methods

"""
import datetime
import typing
from email import utils

import arrow
import iso8601
from dateutil import tz


def age(value: typing.Union[datetime.datetime, str]) -> datetime.timedelta:
    """Return the age of a timestamp as a datetime.timedelta"""
    if isinstance(value, str):
        return utcnow() - iso8601.parse_date(value)
    return utcnow() - value


def isoformat(value: typing.Optional[datetime.datetime] = None) -> str:
    """Format a datetime Object as an ISO-8601 timestamp without milliseconds.
    If the value is not returned, return the current time in UTC.

    """
    if not value:
        value = utcnow()
    output = value.isoformat(' ')
    if '.' in output:
        parts = output.split('.')
        return '{}{}'.format(parts[0], parts[1][6:])
    return output


def parse(value: str) -> datetime.datetime:
    """Parse an ISO-8601 formatted timestamp"""
    return iso8601.parse_date(value)


def parse_rfc822(value: str) -> typing.Optional[datetime.datetime]:
    """Parse an RFC-822 formatted timestamp value, returning a
    :class:`~datetime.datetime` instance.

    """
    parsed = utils.parsedate_tz(value)
    if not parsed:
        return None
    return datetime.datetime.fromtimestamp(utils.mktime_tz(parsed),
                                           tz.tzoffset(None, 0))


def to_utc(value: str) -> str:
    """Convert an ISO-8601 formatted timestamp to UTC"""
    return isoformat(arrow.get(value).to('UTC'))


def utcnow() -> datetime.datetime:
    """Return the current time in UTC"""
    return datetime.datetime.now(tz=tz.tzoffset(None, 0))



import re

from imbi.endpoints import base


class _RequestHandlerMixin:

    ID_KEY = 'id'
    FIELDS = [
        'id', 'name', 'plural_name', 'description', 'icon_class',
        'environment_urls', 'gitlab_project_prefix', 'github_org'
    ]
    DEFAULTS = {'icon_class': 'fas fa-folder'}

    GET_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT id, "name", created_at, created_by,
               last_modified_at, last_modified_by,
               description, plural_name, slug, icon_class,
               environment_urls, gitlab_project_prefix, github_org
          FROM v1.project_types
         WHERE id=%(id)s""")


class CollectionRequestHandler(_RequestHandlerMixin,
                               base.CollectionRequestHandler):

    NAME = 'project-types'
    ITEM_NAME = 'project-type'

    COLLECTION_SQL = re.sub(
        r'\s+', ' ', """\
        SELECT id, "name", plural_name, description, slug, icon_class,
               environment_urls, gitlab_project_prefix, github_org
          FROM v1.project_types
         ORDER BY "name" ASC""")

    POST_SQL = re.sub(
        r'\s+', ' ', """\
        INSERT INTO v1.project_types
                    ("name", created_by, plural_name, description,
                     slug, icon_class, environment_urls, gitlab_project_prefix,
                     github_org)
             VALUES (%(name)s, %(username)s, %(plural_name)s, %(description)s,
                     %(slug)s, %(icon_class)s, %(environment_urls)s,
                     %(gitlab_project_prefix)s, %(github_org)s)
          RETURNING id""")


class RecordRequestHandler(_RequestHandlerMixin, base.AdminCRUDRequestHandler):

    NAME = 'project-type'

    DELETE_SQL = 'DELETE FROM v1.project_types WHERE id=%(id)s;'

    PATCH_SQL = re.sub(
        r'\s+', ' ', """\
        UPDATE v1.project_types
           SET "name"=%(name)s,
               last_modified_at=CURRENT_TIMESTAMP,
               last_modified_by=%(username)s,
               plural_name=%(plural_name)s,
               description=%(description)s,
               slug=%(slug)s,
               icon_class=%(icon_class)s,
               environment_urls=%(environment_urls)s,
               gitlab_project_prefix=%(gitlab_project_prefix)s,
               github_org=%(github_org)s
         WHERE id=%(id)s""")



import datetime
import unittest

from dateutil import tz

from imbi import timestamp


class TimestampTests(unittest.TestCase):
    def test_format_and_parse(self):
        expectation = '2016-08-26 13:46:34-04:00'
        parsed = timestamp.parse(expectation)
        self.assertIsInstance(parsed, datetime.datetime)
        self.assertEqual(expectation, timestamp.isoformat(parsed))

    def test_isoformat_now(self):
        expectation = \
            datetime.datetime.utcnow().isoformat(' ').split('.')[0] + '+00:00'
        self.assertEqual(expectation, timestamp.isoformat())

    def test_parse_rfc822(self):
        expectation = datetime.datetime(1994,
                                        10,
                                        29,
                                        19,
                                        43,
                                        31,
                                        tzinfo=tz.tzoffset(None, 0))
        self.assertEqual(
            expectation,
            timestamp.parse_rfc822('Sat, 29 Oct 1994 19:43:31 GMT'))

    def test_parse_rfc822_invalid(self):
        self.assertIsNone(timestamp.parse_rfc822('Foo Bar'))

    def test_utcnow(self):
        self.assertEqual(
            timestamp.isoformat(timestamp.utcnow()),
            timestamp.isoformat(
                datetime.datetime.now(tz=tz.tzoffset(None, 0))))



import asyncio
import dataclasses
import datetime
import decimal
import logging
import re
import typing

from imbi import common, errors
from imbi.automations.models import (Automation, Integration, automation,
                                     integration)

if typing.TYPE_CHECKING:
    from imbi import app

# chose to use an explicit export list here to simplify
# management of models across different modules
__all__ = [
    'Automation',
    'CookieCutter',
    'Integration',
    'Namespace',
    'OperationsLog',
    'Project',
    'ProjectFact',
    'ProjectIdentifier',
    'ProjectLink',
    'ProjectType',
    'ProjectURL',
    'automation',
    'integration',
    'namespace',
    'operations_log',
    'project',
    'project_facts',
    'project_identifiers',
    'project_links',
    'project_type',
    'project_urls',
]

decimal.getcontext().prec = 2
LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class CookieCutter:
    name: str
    project_type_id: int
    url: str


@dataclasses.dataclass
class Namespace:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: str
    last_modified_by: typing.Optional[datetime.datetime]
    name: str
    slug: str
    icon_class: str
    maintained_by: typing.Optional[list[str]]
    gitlab_group_name: str
    sentry_team_slug: typing.Optional[str]
    pagerduty_policy: typing.Optional[str]
    aws_ssm_slug: typing.Optional[str]

    SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               name,
               slug,
               icon_class,
               maintained_by,
               gitlab_group_name,
               sentry_team_slug,
               pagerduty_policy,
               aws_ssm_slug
          FROM v1.namespaces
         WHERE id=%(id)s""")


@dataclasses.dataclass
class ProjectComponent:
    name: str
    package_url: str
    status: str
    version: str

    COLLECTION_SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT c.name, pc.package_url, c.status, v.version
          FROM v1.project_components AS pc
          JOIN v1.component_versions AS v
            ON pc.package_url = v.package_url AND pc.version_id = v.id
          JOIN v1.components AS c ON c.package_url = pc.package_url
         WHERE pc.project_id = %(obj_id)s
         ORDER BY pc.package_url, v.version
        """)


@dataclasses.dataclass
class ProjectDependency:
    project_id: int
    dependency_id: int
    created_at: datetime.datetime
    created_by: str

    SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
            SELECT project_id, dependency_id, created_at, created_by
              FROM v1.project_dependencies
             WHERE project_id = %(project_id)s
               AND dependency_id = %(dependency_id)s""")

    COLLECTION_SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
            SELECT project_id, dependency_id, created_at, created_by
              FROM v1.project_dependencies
             WHERE project_id = %(obj_id)s
            """)


@dataclasses.dataclass
class ProjectFact:
    id: int
    name: str
    recorded_at: datetime.datetime
    recorded_by: str
    value: typing.Union[bool, datetime.date, datetime.datetime,
                        decimal.Decimal, int, None, str]
    fact_type: str
    data_type: str
    description: typing.Optional[str]
    ui_options: typing.Optional[list[str]]
    score: decimal.Decimal
    weight: int

    COLLECTION_SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        WITH project_type_id AS (SELECT project_type_id AS id
                                   FROM v1.projects
                                  WHERE id = %(obj_id)s)
        SELECT a.id,
               a.name,
               b.recorded_at,
               b.recorded_by,
               b.value,
               a.fact_type,
               a.data_type,
               a.description,
               a.ui_options,
               CASE WHEN b.value IS NULL THEN 0
                    ELSE CASE WHEN a.fact_type = 'enum' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_enums
                                           WHERE fact_type_id = b.fact_type_id
                                             AND value = b.value)
                              WHEN a.fact_type = 'range' THEN (
                                          SELECT score::NUMERIC(9,2)
                                            FROM v1.project_fact_type_ranges
                                           WHERE fact_type_id = b.fact_type_id
                                             AND b.value::NUMERIC(9,2)
                                         BETWEEN min_value AND max_value)
                              ELSE 0
                          END
                END AS score,
               a.weight
          FROM v1.project_fact_types AS a
     LEFT JOIN v1.project_facts AS b
            ON b.fact_type_id = a.id
           AND b.project_id = %(obj_id)s
         WHERE (SELECT id FROM project_type_id) = ANY(a.project_type_ids)
      ORDER BY a.name""")

    def __post_init__(self):
        try:
            value = common.coerce_project_fact(self.data_type, self.value)
        except ValueError:
            pass
        else:
            self.__setattr__('value', value)


@dataclasses.dataclass
class ProjectIdentifier:
    integration_name: str
    external_id: str
    created_at: datetime.datetime
    created_by: str

    COLLECTION_SQL: typing.ClassVar[str] = re.sub(
        r'\s+', ' ', """\
        SELECT integration_name, external_id, created_at, created_by
          FROM v1.project_identifiers
         WHERE project_id = %(obj_id)s
         ORDER BY integration_name""")


@dataclasses.dataclass
class ProjectType:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    name: str
    slug: str
    plural_name: str
    description: typing.Optional[str]
    icon_class: typing.Optional[str]
    environment_urls: bool
    gitlab_project_prefix: typing.Optional[str]
    github_org: typing.Optional[str]

    SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               name,
               slug,
               plural_name,
               description,
               icon_class,
               environment_urls,
               gitlab_project_prefix,
               github_org
          FROM v1.project_types
         WHERE id=%(id)s""")


@dataclasses.dataclass
class ProjectLink:
    link_type_id: int
    link_type: str
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    icon_class: typing.Optional[str]
    url: str

    COLLECTION_SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT a.link_type_id,
               b.link_type,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.icon_class,
               a.url
          FROM v1.project_links AS a
          JOIN v1.project_link_types AS b
            ON b.id = a.link_type_id
         WHERE a.project_id = %(obj_id)s
         ORDER BY b.link_type""")


@dataclasses.dataclass
class ProjectURL:
    environment: str
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[str]
    last_modified_by: typing.Optional[datetime.datetime]
    icon_class: typing.Optional[str]
    url: str

    COLLECTION_SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT a.environment,
               a.created_at,
               a.created_by,
               a.last_modified_at,
               a.last_modified_by,
               b.icon_class,
               a.url
          FROM v1.project_urls AS a
          JOIN v1.environments AS b
            ON b.name = a.environment
         WHERE a.project_id = %(obj_id)s
         ORDER BY b.name""")


@dataclasses.dataclass
class Project:
    id: int
    created_at: datetime.datetime
    created_by: str
    last_modified_at: typing.Optional[datetime.datetime]
    last_modified_by: typing.Optional[str]
    namespace: Namespace
    project_type: ProjectType
    name: str
    slug: str
    description: typing.Optional[str]
    environments: typing.Optional[list[str]]
    archived: bool
    gitlab_project_id: typing.Optional[int]
    sentry_project_slug: typing.Optional[str]
    sonarqube_project_key: typing.Optional[str]
    pagerduty_service_id: typing.Optional[str]
    facts: dict[str, str]
    identifiers: dict[str, typing.Any]
    links: dict[str, str]
    urls: dict[str, str]
    project_score: int
    components: list[dict[str, str]]
    component_versions: list[str]
    dependencies: typing.Optional[list[int]]

    SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT id,
               created_at,
               created_by,
               last_modified_at,
               last_modified_by,
               namespace_id,
               project_type_id,
               name,
               slug,
               description,
               environments,
               archived,
               gitlab_project_id,
               sentry_project_slug,
               sonarqube_project_key,
               pagerduty_service_id,
               v1.project_score(id) AS project_score
          FROM v1.projects
         WHERE id=%(id)s""")


@dataclasses.dataclass
class OperationsLog:
    id: int
    occurred_at: datetime.datetime
    recorded_at: datetime.datetime
    recorded_by: str
    display_name: str
    completed_at: typing.Optional[datetime.datetime]
    project_id: typing.Optional[int]
    project_name: typing.Optional[str]
    environment: str
    change_type: str
    description: typing.Optional[str]
    link: typing.Optional[str]
    notes: typing.Optional[str]
    performed_by: typing.Optional[str]
    ticket_slug: typing.Optional[str]
    version: typing.Optional[str]

    SQL: typing.ClassVar = re.sub(
        r'\s+', ' ', """\
        SELECT o.id,
               o.occurred_at,
               o.recorded_at,
               o.recorded_by,
               COALESCE(
                  u.display_name,
                  COALESCE(u2.display_name, o.recorded_by)
               ) AS display_name,
               o.completed_at,
               o.project_id,
               p.name AS project_name,
               o.environment,
               o.change_type,
               o.description,
               o.link,
               o.notes,
               o.performed_by,
               o.ticket_slug,
               o.version
          FROM v1.operations_log AS o
     LEFT JOIN v1.users AS u
            ON u.username = o.performed_by
     LEFT JOIN v1.users AS u2
            ON u2.username = o.recorded_by
     LEFT JOIN v1.projects AS p
            ON o.project_id = p.id
         WHERE o.id=%(id)s""")


async def _load(model: dataclasses.dataclass, application: 'app.Application',
                **params) -> dataclasses.dataclass:
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for %s with params %r: %s',
                     model.__class__.__name__, params, exc)
        raise errors.DatabaseError(f'Error loading {model.__class__.__name__}',
                                   error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(model.SQL, params, 'model-load')
        if result.row_count:
            return model(**result.row)


async def _load_collection(model: dataclasses.dataclass,
                           obj_id: int,
                           application: 'app.Application') \
        -> list[dataclasses.dataclass]:
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for collection %s: %s', obj_id,
                     exc)
        raise errors.DatabaseError(
            f'Error loading {model.__class__.__name__} '
            f'collection for {obj_id}',
            error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(model.COLLECTION_SQL, {'obj_id': obj_id},
                                    'collection-load')
        return [model(**row) for row in result.rows]


async def namespace(namespace_id: int,
                    application: 'app.Application') -> Namespace:
    return await _load(Namespace, application, id=namespace_id)


async def operations_log(ops_log_id: int,
                         application: 'app.Application') -> OperationsLog:
    return await _load(OperationsLog, application, id=ops_log_id)


async def project(project_id: int, application: 'app.Application') -> Project:
    def on_postgres_error(_metric_name: str, exc: Exception) -> None:
        LOGGER.error('Failed to execute query for project %s: %s', project_id,
                     exc)
        raise errors.DatabaseError(f'Error loading Project {project_id}',
                                   error=exc)

    async with application.postgres_connector(
            on_error=on_postgres_error) as conn:
        result = await conn.execute(Project.SQL, {'id': project_id},
                                    'project-model-load')
        if result.row_count:
            values = dict(result.row)
            results: tuple[Namespace, ProjectType, list[ProjectFact],
                           list[ProjectLink], list[ProjectURL],
                           list[ProjectIdentifier], list[ProjectComponent],
                           list[ProjectDependency]]
            results = await asyncio.gather(
                namespace(values['namespace_id'], application),
                project_type(values['project_type_id'], application),
                project_facts(project_id, application),
                project_links(project_id, application),
                project_urls(project_id, application),
                project_identifiers(project_id, application),
                project_components(project_id, application),
                project_dependencies(project_id, application),
            )
            del values['namespace_id']
            del values['project_type_id']
            values.update({
                'namespace': results[0],
                'project_type': results[1],
                'facts': {value.name: value.value
                          for value in results[2]},
                'links': {value.link_type: value.url
                          for value in results[3]},
                'urls': {value.environment: value.url
                         for value in results[4]},
                'identifiers': {
                    value.integration_name: value.external_id
                    for value in results[5]
                },
                'components': [{
                    'name': c.package_url,
                    'version': c.version
                } for c in results[6]],
                'component_versions': [
                    f'{c.package_url}@{c.version}' for c in results[6]
                ],
                'dependencies': [d.dependency_id for d in results[7]]
            })
            return Project(**values)


async def project_components(
        project_id: int,
        application: 'app.Application') -> list[ProjectComponent]:
    return await _load_collection(ProjectComponent, project_id, application)


async def project_dependency(
        project_id: int, dependency_id: int,
        application: 'app.Application') -> ProjectDependency:
    return await _load(ProjectDependency,
                       application,
                       project_id=project_id,
                       dependency_id=dependency_id)


async def project_dependencies(
        project_id: int,
        application: 'app.Application') -> list[ProjectDependency]:
    return await _load_collection(ProjectDependency, project_id, application)


async def project_facts(project_id: int,
                        application: 'app.Application') \
        -> list[ProjectFact]:
    return await _load_collection(ProjectFact, project_id, application)


async def project_identifiers(
        project_id: int,
        application: 'app.Application') -> list[ProjectIdentifier]:
    return await _load_collection(ProjectIdentifier, project_id, application)


async def project_links(project_id: int,
                        application: 'app.Application') \
        -> list[ProjectLink]:
    return await _load_collection(ProjectLink, project_id, application)


async def project_type(project_type_id: int,
                       application: 'app.Application') -> ProjectType:
    return await _load(ProjectType, application, id=project_type_id)


async def project_urls(project_id: int,
                       application: 'app.Application') \
        -> list[ProjectURL]:
    return await _load_collection(ProjectURL, project_id, application)



"""
Modules for interacting with and materializing data for OpenSearch Indexes
"""
from __future__ import annotations

import asyncio
import contextlib
import logging
import typing

import sprockets.mixins.mediatype.content
import sprockets_postgres
import tornado.web

from imbi import errors
if typing.TYPE_CHECKING:
    from imbi import app

ModelType = typing.TypeVar('ModelType')


class SupportsIndexDocumentById(typing.Protocol):
    async def index_document_by_id(self, doc_id: int) -> bool:
        ...


class SearchIndex(typing.Generic[ModelType]):
    INDEX: typing.ClassVar[str] = ''

    def __init__(
        self, application: 'app.Application',
        fetch_method: typing.Callable[[int, 'app.Application'],
                                      typing.Awaitable[ModelType]]
    ) -> None:
        super().__init__()
        self.application = application
        self.fetch_document_by_id = fetch_method
        self.logger = logging.getLogger(__name__).getChild(
            self.__class__.__name__)

    async def create_index(self) -> None:
        await self.application.opensearch.create_index(self.INDEX)

    async def create_mapping(self) -> None:
        await self.application.opensearch.create_mapping(
            self.INDEX, await self._build_mappings())

    async def delete_document(self, doc_id: int | str) -> None:
        await self.application.opensearch.delete_document(
            self.INDEX, str(doc_id))

    async def index_document(self, doc: ModelType) -> None:
        await self.application.opensearch.index_document(
            self.INDEX, str(doc.id), self._serialize_document(doc), True)

    async def index_document_by_id(self, doc_id: int) -> bool:
        doc = None
        with contextlib.suppress(errors.DatabaseError):
            doc = await self.fetch_document_by_id(doc_id, self.application)
        if doc is None:
            self.logger.warning('Document not found for %s while indexing',
                                doc_id)
            return False
        await self.index_document(doc)
        return True

    async def search(self,
                     query: str,
                     max_results: int = 1000) -> dict[str, list[dict]]:
        return await self.application.opensearch.search(
            self.INDEX, query, max_results)

    async def _build_mappings(self):
        raise NotImplementedError()

    def _serialize_document(self, doc: ModelType) -> dict[str, typing.Any]:
        raise NotImplementedError()


class SearchIndexRequestHandler(
        sprockets_postgres.RequestHandlerMixin,
        sprockets.mixins.mediatype.content.ContentMixin,
        tornado.web.RequestHandler,
):
    search_index: SupportsIndexDocumentById
    SQL: typing.ClassVar[str] = ''

    async def prepare(self) -> None:
        maybe_coro = super().prepare()
        if asyncio.isfuture(maybe_coro) or asyncio.iscoroutine(maybe_coro):
            await maybe_coro
        if not self.SQL:
            raise errors.InternalServerError(
                'Programming error in %s: SQL is not defined',
                self.__class__.__name__,
                reason='Internal Server Error',
                title='Programming Error',
                detail='SQL is not defined',
                source=self.__class__.__name__,
            )
        if getattr(self, 'search_index', None) is None:
            raise errors.InternalServerError(
                'Programming error in %s: search_index is None',
                self.__class__.__name__,
                reason='Internal Server Error',
                title='Programming Error',
                detail='search_index is None',
                source=self.__class__.__name__,
            )

    async def post(self) -> None:
        docs_to_index: list[int] = []
        if ids := self.get_query_arguments('id'):
            try:
                docs_to_index.extend(int(arg) for arg in ids)
            except ValueError:
                raise errors.BadRequest('Invalid ID found in %r', ids)
        else:
            docs_to_index.extend(
                r['id'] for r in await self.postgres_execute(self.SQL))

        indexed_documents = 0
        for doc_id in docs_to_index:
            if await self.search_index.index_document_by_id(doc_id):
                indexed_documents += 1

        self.send_response({
            'status': 'ok',
            'message': f'Queued {indexed_documents} documents for indexing'
        })



from tornado import testing

from imbi import models
from tests import base


class NamespaceTestCase(base.TestCaseWithReset):

    TRUNCATE_TABLES = ['v1.namespaces']

    def setUp(self) -> None:
        super().setUp()
        self.namespace = self.create_namespace()

    @testing.gen_test
    async def test_namespace_model(self):
        namespace = await models.namespace(self.namespace['id'], self._app)
        self.assertEqual(namespace.name, self.namespace['name'])
        self.assertEqual(namespace.slug, self.namespace['slug'])
        self.assertEqual(namespace.icon_class, self.namespace['icon_class'])


class ProjectTypeTestCase(base.TestCaseWithReset):

    TRUNCATE_TABLES = ['v1.project_types']

    def setUp(self) -> None:
        super().setUp()
        self.project_type = self.create_project_type()

    @testing.gen_test
    async def test_namespace_model(self):
        value = await models.project_type(self.project_type['id'], self._app)
        self.assertEqual(value.name, self.project_type['name'])
        self.assertEqual(value.slug, self.project_type['slug'])
        self.assertEqual(value.icon_class, self.project_type['icon_class'])
